# A TypedEnum can be used to represent a typed, enumerated set of values.  A
# TypedEnum offers the following features:
#
#  * Instances are restricted (in the absence of class-opening) to the set of instances defined in
#    the class definition
#  * Available values can be enumerated in a consistent order via a call to the 
#  * #values class method
#  * Instances can be retrieved via class constants
#  * Instances can be retrieved by parsing a string/symbol
#  * Instances can be serialized to a string
#
# Basic usage example:
#   class Color
#     @enum = lambda {[
#       RED   = self.new(:red),
#       GREEN = self.new(:green),
#       BLUE  = self.new(:blue)
#     ]}
#     include TypedEnum
#   end
#
#   Color::RED => #<Color:0x163d59f8 @name=:red>
#   Color.from_name(:red) => #<Color:0x163d59f8 @name=:red>
#   Color.from_name('red') => #<Color:0x163d59f8 @name=:red>
#   Color::RED.name => :red
#   Color.values => [#<Color:0x163d59f8 @name=:red>, #<Color:0x163d59a8 @name=:green>, #<Color:0x163d5980 @name=:blue>]
#   
module TypedEnum

  attr_reader :name
  def initialize(name)
    @name = name
  end

  def clone
    raise TypeError, "Can't clone instance of TypedEnum #{self.class}"
  end
  def dup
    raise TypeError, "Can't dup instance of TypedEnum #{self.class}"
  end
  def dump(depth=-1)
    @name.to_s
  end
end

class << TypedEnum
  module TypedEnumClassMethods
    def from_name(sym)
      inverse_map[sym.to_sym] or raise NameError.new("No mapping for symbol #{sym} to vertical")
    end

    def parse(token)
      sym = token.to_sym
      from_symbol(sym)
    end

    def values
      @@values
    end

    def _load(str)
      parse(str.to_sym)
    end

    @initialized = false
    def init
      if !@initialized
        @initialized = true
        @@values = @enum.call
      else
        raise "Attempt to re-initialize typed enum #{self.class}"
      end
    end

    private

    def inverse_map
      @@inverse_map ||= values.inject({}) { |m, v| m[v.name] = v; m }
    end

  end

  def TypedEnum.included(klass)
    super
    klass.extend TypedEnumClassMethods
    klass.init
    klass.private_class_method :new, :allocate
  end
end

if __FILE__ == $0

  class Foo
    @enum = lambda { [BOOKS = self.new(:books), GAMES = self.new(:games)] }
    include TypedEnum
  end

  require 'pp'

  pp (Foo.methods - Object.methods).sort

  pp Foo.values

end
